"""
This script creates a "Prez Resources" table in either Markdown or ASCIIDOC from a Manifest file which it validates first

Run this script with the -h flag for more help, i.e. ~$ python documentor.py -h

Example:

Input:

PREFIX mrr: <https://prez.dev/ManifestResourceRoles/>
PREFIX prez: <https://prez.dev/>
PREFIX prof: <http://www.w3.org/ns/dx/prof/>
PREFIX schema: <https://schema.org/>

[]
    a prez:Manifest ;
    prof:hasResource
        [
            prof:hasArtifact "catalogue.ttl" ;
            prof:hasRole mrr:ContainerData ;
            schema:description "The definition of, and medata for, the container which here is a dcat:Catalog object" ;
            schema:name "Catalogue Definition"
        ] ,
        [
            prof:hasArtifact "vocabs/*.ttl" ;
            prof:hasRole mrr:ContentData ;
            schema:description "skos:ConceptScheme objects in RDF (Turtle) files in the vocabs/ folder" ;
            schema:name "Content"
        ] ,
        [
            prof:hasArtifact "https://github.com/RDFLib/prez/blob/main/prez/reference_data/profiles/ogc_records_profile.ttl" ;
            prof:hasRole mrr:ContainerAndContentModel ;
            schema:description "The default Prez profile for Records API" ;
            schema:name "Profile Definition"
        ] ,
        [
            prof:hasArtifact "_background/labels.ttl" ;
            prof:hasRole mrr:CompleteContainerAndContentLabels ;
            schema:description "An RDF file containing all the labels for the container content" ;
        ] ;
    .

Output:

Resource | Role | Description
--- | --- | ---
Catalogue Definition, [`catalogue.ttl`](catalogue.ttl) | [Container Data](https://prez.dev/ManifestResourceRoles/ContainerData) | The definition of, and medata for, the container which here is a dcat:Catalog object
Content, [`vocabs/*.ttl`](vocabs/*.ttl) | [Content Data](https://prez.dev/ManifestResourceRoles/ContentData) | skos:ConceptScheme objects in RDF (Turtle) files in the vocabs/ folder
Profile Definition, [`ogc_records_profile.ttl`](https://github.com/RDFLib/prez/blob/main/prez/reference_data/profiles/ogc_records_profile.ttl) | [Container & Content Model](https://prez.dev/ManifestResourceRoles/ContainerAndContentModel) | The default Prez profile for Records API
Labels file, [`_background/labels.ttl`](_background/labels.ttl) | [Complete Content and Container Labels](https://prez.dev/ManifestResourceRoles/CompleteContainerAndContentLabels) | An RDF file containing all the labels for the container content
"""

import argparse
import sys
from pathlib import Path
from urllib.parse import ParseResult, urlparse

from kurra.utils import load_graph
from rdflib import DCAT, DCTERMS, PROF, SDO

try:
    from prezmanifest import MRR, validate, __version__
    from prezmanifest.utils import *
except ImportError:
    import sys

    sys.path.append(str(Path(__file__).parent.parent.resolve()))
    from prezmanifest import MRR, validate, __version__
    from prezmanifest.utils import *


def create_table(manifest: Path, t="markdown") -> str:
    # load and validate manifest
    validate(manifest)
    manifest_graph = load_graph(manifest)

    # add in MRR vocab
    manifest_graph += load_graph(Path(__file__).parent / "mrr.ttl")

    if t == "asciidoc":
        header = "|===\n| Resource | Role | Description\n\n"
    else:
        header = "Resource | Role | Description\n--- | --- | ---\n"

    body = ""
    for s, o in manifest_graph.subject_objects(PROF.hasResource):
        a = str(manifest_graph.value(o, PROF.hasArtifact))
        if t == "asciidoc":
            artifact = f'link:{a}[`{a.split("/")[-1] if a.startswith("http") else a}`]'
        else:
            artifact = f'[`{a.split("/")[-1] if a.startswith("http") else a}`]({a})'
        role_iri = manifest_graph.value(o, PROF.hasRole)
        role_label = manifest_graph.value(role_iri, SKOS.prefLabel)
        if t == "asciidoc":
            role = f"{role_iri}[{role_label}]"
        else:
            role = f"[{role_label}]({role_iri})"
        name = manifest_graph.value(o, SDO.name)
        description = manifest_graph.value(o, SDO.description)
        n = f"{name}, {artifact}" if name is not None else f"{artifact}"
        d = description if description is not None else ""
        if t == "asciidoc":
            body += f"| {n} | {role} | {d}\n"
        else:
            body += f"{n} | {role} | {d}\n"

    if t == "asciidoc":
        footer = "|===\n"
    else:
        footer = ""

    return (header + body + footer).strip()


def create_catalogue(manifest: Path):
    MANIFEST_ROOT_DIR = manifest.parent
    # load and validate manifest
    validate(manifest)
    manifest_graph = load_graph(manifest)

    catalogue = Graph()
    for s, o in manifest_graph.subject_objects(PROF.hasResource):
        for role in manifest_graph.objects(o, PROF.hasRole):
            if role == MRR.CatalogueData:
                for artifact in manifest_graph.objects(o, PROF.hasArtifact):
                    # the artifact can only be a triples file (not a quads file)
                    catalogue = load_graph(MANIFEST_ROOT_DIR / artifact)

    # get the IRI of the catalogue
    catalogue_iri = catalogue.value(predicate=RDF.type, object=DCAT.Catalog) or \
                    catalogue.value(predicate=RDF.type, object=SDO.DataCatalog)

    # non-catalogue resources
    for s, o in manifest_graph.subject_objects(PROF.hasResource):
        for role in manifest_graph.objects(o, PROF.hasRole):
            if role == MRR.ResourceData:
                for artifact in manifest_graph.objects(o, PROF.hasArtifact):
                    for f in get_files_from_artifact(manifest, artifact):
                        for iri in sorted(get_identifier_from_file(f)):
                            if iri != URIRef("urn:x-rdflib:default"):
                                catalogue.add((catalogue_iri, SDO.hasPart, iri))

    return catalogue


def setup_cli_parser(args=None):
    def url_file_or_folder(input: str) -> ParseResult | Path:
        parsed = urlparse(input)
        if all([parsed.scheme, parsed.netloc]):
            return parsed
        path = Path(input)
        if path.is_file():
            return path
        if path.is_dir():
            return path
        raise argparse.ArgumentTypeError(
            f"{input} is not a valid input. Must be a file, folder or sparql endpoint"
        )

    parser = argparse.ArgumentParser()

    parser.add_argument(
        "-v",
        "--version",
        action="version",
        version="{version}".format(version=__version__),
    )

    parser.add_argument(
        "function",
        help="The documentation function you wish to perform",
        choices=["table", "catalogue"],
    )

    parser.add_argument(
        "-t",
        "--type",
        help="The type of markup you want to export: Markdown or ASCCIDOC. Only relevant for the 'table' function",
        choices=["markdown", "asciidoc"],
        default="markdown",
    )

    parser.add_argument(
        "input",
        help="A Prez Manifest file",
        type=url_file_or_folder,
    )

    return parser.parse_args(args)


def cli(args=None):
    if args is None:
        args = sys.argv[1:]

    args = setup_cli_parser(args)

    if args.function == "table":
        print(create_table(args.input, t=args.type))
    else:
        print(create_catalogue(args.input).serialize(format="longturtle"))


if __name__ == "__main__":
    retval = cli(sys.argv[1:])
    if retval is not None:
        sys.exit(retval)



from pathlib import Path

from rdflib import Graph
from rdflib.namespace import RDF, SKOS

FILES_DIR = Path("something")

for f in FILES_DIR.glob("*.ttl"):
    g = Graph()
    g.parse(f)
    iri = g.value(predicate=RDF.type, object=SKOS.ConceptScheme)

    # print(f"{f.name}, {iri}")

    print(f"<{iri}> ,")



"""
Assesses a given Manifest, finds any IRIs in any of the given resources missing labels and tries to patch them from
a given source of labels, such as KurrawongAI's Semantic Background (https://github.com/kurrawong/semantic-background)
repository.
"""

import argparse
import sys
from pathlib import Path
from typing import Literal as TLiteral
from urllib.parse import ParseResult, urlparse

from kurra.utils import load_graph
from labelify import find_missing_labels, extract_labels
from rdflib import Graph, BNode, Literal, URIRef
from rdflib.namespace import PROF, RDFS

from prezmanifest.utils import get_files_from_artifact

try:
    from prezmanifest import MRR, OLIS, validate, load, __version__
except ImportError:
    import sys

    sys.path.append(str(Path(__file__).parent.parent.resolve()))
    from prezmanifest import MRR, OLIS, validate, __version__


def label(
    manifest: Path,
    output: TLiteral["iris", "rdf", "manifest"] = "manifest",
    additional_context: Path | str | Graph = None,
) -> set | Graph | None:
    """ "Main function for labeller module"""
    # create the target from the Manifest
    manifest_content_graph = load(manifest, return_data_type="Graph")

    output_types = ["iris", "rdf", "manifest"]
    if output not in output_types:
        raise ValueError(
            f"Parameter output is {output} but must be one of {', '.join(output_types)}"
        )

    # determine if any labelling context is given in Manifest
    context_graph = Graph()
    for s, o in manifest_content_graph.subject_objects(PROF.hasResource):
        for role in manifest_content_graph.objects(o, PROF.hasRole):
            if role in [
                MRR.IncompleteCatalogueAndResourceLabels,
                MRR.CompleteCatalogueAndResourceLabels,
            ]:
                for artifact in manifest_content_graph.objects(o, PROF.hasArtifact):
                    artifact: Literal
                    for f in get_files_from_artifact(manifest, artifact):
                        context_graph += load_graph(f)

    # add labels for system IRIs
    context_graph.parse(Path(__file__).parent / "system-labels.ttl")


    if output == "iris":
        return find_missing_labels(
            manifest_content_graph + context_graph, additional_context
        )

    elif output == "rdf":
        iris = find_missing_labels(manifest_content_graph, context_graph)

        if additional_context is not None:
            return extract_labels(iris, additional_context)
        else:
            return None

    else:  # output == manifest
        # If this is selected, generate the "rdf" output and create a resource for it in the Manifest
        # If there are no more missing labels then we have an mrr:CompleteCatalogueAndResourceLabels
        # else add mrr:IncompleteCatalogueAndResourceLabels

        # Generate labels for any IRIs missing them, using context given in the Manifest and any
        # Additional Context supplied
        manifest_only_graph = load_graph(manifest)
        rdf_addition = label(manifest, "rdf", additional_context)

        if len(rdf_addition) > 0:
            new_artifact = manifest.parent / "labels-additional.ttl"
            rdf_addition.serialize(destination=new_artifact, format="longturtle")
            new_resource = BNode()

            # Find the role of any context in the Manifest
            manifest_iri = None
            context_roles = []
            for s, o in manifest_only_graph.subject_objects(PROF.hasResource):
                manifest_iri = s
                for role in manifest_only_graph.objects(o, PROF.hasRole):
                    if role in [
                        MRR.IncompleteCatalogueAndResourceLabels,
                        MRR.CompleteCatalogueAndResourceLabels,
                    ]:
                        context_roles.append(role)

            if (
                MRR.CompleteCatalogueAndResourceLabels in context_roles
                and len(context_roles) == 1
            ):
                # If a CompleteCatalogueAndResourceLabels is present in Manifest and yet more labels were discovered,
                # change CompleteCatalogueAndResourceLabels to IncompleteCatalogueAndResourceLabels and add another
                for s, o in manifest_content_graph.subject_objects(PROF.hasRole):
                    if o == MRR.CompleteCatalogueAndResourceLabels:
                        manifest_only_graph.remove((s, PROF.hasRole, o))
                        manifest_only_graph.add(
                            (manifest_iri, PROF.hasResource, new_resource)
                        )
                        manifest_only_graph.add(
                            (
                                new_resource,
                                PROF.hasRole,
                                MRR.IncompleteCatalogueAndResourceLabels,
                            )
                        )
                        manifest_only_graph.add(
                            (new_resource, PROF.hasArtifact, Literal(new_artifact.name))
                        )
            else:
                # If an IncompleteCatalogueAndResourceLabels was present, add another IncompleteCatalogueAndResourceLabels
                # which together make a CompleteCatalogueAndResourceLabels

                # If none was present, add an IncompleteCatalogueAndResourceLabels or a CompleteCatalogueAndResourceLabels
                # TODO: test for completeness of labelling and add in CompleteCatalogueAndResourceLabels if complete
                manifest_only_graph.add((manifest_iri, PROF.hasResource, new_resource))
                manifest_only_graph.add(
                    (
                        new_resource,
                        PROF.hasRole,
                        MRR.IncompleteCatalogueAndResourceLabels,
                    )
                )
                manifest_only_graph.add(
                    (new_resource, PROF.hasArtifact, Literal(new_artifact.name))
                )

            manifest_only_graph.serialize(destination=manifest, format="longturtle")

        else:
            raise Warning(
                "No new labels have been generated for content in this Manifest. "
                "This could be because none were missing or because no new labels can be found in any "
                "supplied additional context."
            )


def setup_cli_parser(args=None):
    def url_file_or_folder(input: str) -> ParseResult | Path:
        parsed = urlparse(input)
        if all([parsed.scheme, parsed.netloc]):
            return parsed
        path = Path(input)
        if path.is_file():
            return path
        if path.is_dir():
            return path
        raise argparse.ArgumentTypeError(
            f"{input} is not a valid input. Must be a file, folder or sparql endpoint"
        )

    parser = argparse.ArgumentParser()
    group = parser.add_mutually_exclusive_group(required=True)

    parser.add_argument(
        "-v",
        "--version",
        action="version",
        version="{version}".format(version=__version__),
    )

    group.add_argument(
        "-o",
        "--output",
        help="The form of output you want",
        choices=["iris", "rdf", "manifest"],
        default="manifest",
    )

    parser.add_argument(
        "-a",
        "--additional-context",
        help="File, Folder or Sparql Endpoint to read additional context RDF from",
        type=url_file_or_folder,
    )

    parser.add_argument(
        "manifest",
        help="A Manifest file to process",
        type=Path,
    )

    return parser.parse_args(args)


def cli(args=None):
    if args is None:
        args = sys.argv[1:]

    args = setup_cli_parser(args)

    x = label(args.manifest, args.output, args.additional_context)

    if args.output == "iris":
        print("\n".join([str(iri) for iri in x]))
    elif args.output == "rdf":
        if x is not None:
            print(x.serialize(format="longturtle"))

    else:  # manifest
        pass


if __name__ == "__main__":
    retval = cli(sys.argv[1:])
    if retval is not None:
        sys.exit(retval)



import importlib.metadata

__version__ = importlib.metadata.version(__package__)

from .definednamespaces import MRR, OLIS, PREZ
from .validator import validate
from .documentor import create_table, create_catalogue
from .loader import load
from .labeller import label



"""Validate a Prez Manifest file.

This script performs both SHACL validation to ensure the Manifest is valid according to the Prez Manifest
specification (see https://prez.dev/manifest/) and checks that all the resources indicated by the Manifest
- whether local files/folders or remote resources on the Internet - are reachable.

~$ python validate.py {MANIFEST_FILE_PATH}"""

import argparse
import sys
from pathlib import Path

import httpx
from kurra.utils import load_graph
from pyshacl import validate as shacl_validate
from rdflib.namespace import PROF

try:
    from prezmanifest import __version__
except ImportError:
    import sys

    sys.path.append(str(Path(__file__).parent.parent.resolve()))
    from prezmanifest import __version__


def validate(manifest: Path) -> bool:
    ME = Path(__file__)
    MANIFEST_ROOT_DIR = manifest.parent

    # SHACL validation
    manifest_graph = load_graph(manifest)
    mrr_vocab_graph = load_graph(ME.parent / "mrr.ttl")
    data_graph = manifest_graph + mrr_vocab_graph
    shacl_graph = load_graph(ME.parent / "validator.ttl")
    valid, v_graph, v_text = shacl_validate(data_graph, shacl_graph=shacl_graph)

    if not valid:
        raise ValueError(f"SHACL invalid:\n\n{v_text}")

    # Content link validation
    for s, o in manifest_graph.subject_objects(PROF.hasResource):
        for artifact in manifest_graph.objects(o, PROF.hasArtifact):
            artifact_str = str(artifact)
            if "http" in artifact_str:
                r = httpx.get(artifact_str)
                if 200 <= r.status_code < 400:
                    pass
                else:
                    raise ValueError(
                        f"Remote content link non-resolving: {artifact_str}"
                    )
            elif "*" in artifact_str:
                glob_parts = artifact_str.split("*")
                dir = Path(manifest.parent / Path(glob_parts[0]))
                if not Path(dir).is_dir():
                    raise ValueError(
                        f"The content link {artifact_str} is not a directory"
                    )
            else:
                # It must be a local
                if not (MANIFEST_ROOT_DIR / artifact_str).is_file():
                    print(
                        f"Content link {MANIFEST_ROOT_DIR / artifact_str} is invalid - not a file"
                    )

    return manifest_graph


def setup_cli_parser(args=None):

    parser = argparse.ArgumentParser()

    parser.add_argument(
        "-v",
        "--version",
        action="version",
        version="{version}".format(version=__version__),
    )

    parser.add_argument(
        "manifest",
        help="A Manifest file to process",
        type=Path,
    )

    return parser.parse_args(args)


def cli(args=None):
    if args is None:
        args = sys.argv[1:]

    args = setup_cli_parser(args)

    validate(args.manifest)


if __name__ == "__main__":
    retval = cli(sys.argv[1:])
    if retval is not None:
        sys.exit(retval)



from collections.abc import Generator
from pathlib import Path
from typing import List

from rdflib import Literal, URIRef, Graph, Dataset
from rdflib.namespace import DCAT, OWL, RDF, SDO, SKOS


def get_files_from_artifact(
    manifest: Path, artifact: Literal
) -> List[Path] | Generator[Path]:
    """Returns an iterable (list or generator) of Path objects for files within an artifact literal.

    This function will correctly interpret artifacts such as 'file.ttl', '*.ttl', '**/*.trig' etc.
    """
    if not "*" in str(artifact):
        return [manifest.parent / Path(str(artifact))]
    else:
        artifact_str = str(artifact)
        glob_marker_location = artifact_str.find("*")
        glob_parts = [
            artifact_str[:glob_marker_location],
            artifact_str[glob_marker_location:],
        ]

        return Path(manifest.parent / Path(glob_parts[0])).glob(glob_parts[1])


def get_identifier_from_file(file: Path) -> List[URIRef]:
    """Returns a list if RDFLib graph identifier (URIRefs) from a triples or quads file
    for all owl:Ontology and skos:ConceptScheme objects"""
    if file.name.endswith(".ttl"):
        g = Graph().parse(file)
        return [
            g.value(predicate=RDF.type, object=SKOS.ConceptScheme)
            or g.value(predicate=RDF.type, object=OWL.Ontology)
            or g.value(predicate=RDF.type, object=DCAT.Resource)
            or g.value(predicate=RDF.type, object=SDO.CreativeWork)
            or g.value(predicate=RDF.type, object=SDO.Dataset)
        ]
    elif file.name.endswith(".trig"):
        gs = []
        d = Dataset()
        d.parse(file, format="trig")
        for g in d.graphs():
            gs.append(g.identifier)
        return gs
    else:
        return []
